# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Version information for a repository.

This module provides the
[`RepoVersionInfo`][frequenz.repo.config.version.RepoVersionInfo] class to get information
about the repository version.

It handles many scenarios and queries, like:

* Is the current commit a tag or a branch?
* Is the current tag the latest tag?
* Is the current branch the latest branch?
* Is the current tag the last minor version for the major version?
* What is the next minor version for the current major branch?
* etc.

Repository tag names are expected to follow the [semantic versioning][semver]
specification, but usually with a leading `v` (e.g. `v1.0.0`).
[`to_semver()`][frequenz.repo.config.version.to_semver] can be used to convert a version
string to a semantic version, even if it has a leading `v`.

Repository branch names can be parsed with
[`BranchVersion.parse()`][frequenz.repo.config.version.BranchVersion.parse] and are
expected follow the format:

- `vX.x.x` for major branches, where `X` is the major version number. For example,
  `v1.x.x` is the major branch for the major version 1.

  It represents an in-development major version, from which new minor branches
  including new features are created (including the first minor branch for the major
  version).

- `vX.Y.x` for minor branches, where `X` is the major version number and `Y` is the
  minor version number. For example, `v1.0.x` is the minor branch for the major version
  1 and minor version 0.

  It represents a maintained minor version, from which new patch releases (with bug
  fixes) are created. For example, `v1.1.x` is the minor branch for the major version 1
  and minor version 1.

[semver]: https://semver.org/
"""

from __future__ import annotations

import dataclasses
import logging
import pathlib
import re
from typing import Self, TypeGuard

import semver

_logger = logging.getLogger(__name__)
_major_branch_re: re.Pattern[str] = re.compile(r"^v?(\d+)\.x\.x$")
_minor_branch_re: re.Pattern[str] = re.compile(r"^v?(\d+)\.(\d+)\.x$")


def strip_v(version: str) -> str:
    """Strip the leading 'v' from a version string.

    Args:
        version: The version string.

    Returns:
        The version string without the leading 'v'.
    """
    return version[1:] if version.startswith("v") else version


def to_semver(version: str) -> semver.Version | None:
    """Convert a version string to a semantic version.

    Args:
        version: The version string.

    Returns:
        The semantic version or `None` if the version string is not a valid semantic
            version.
    """
    try:
        return semver.Version.parse(strip_v(version))
    except ValueError:
        logging.debug("Version is not a semantic version: %s", version)
        return None


def _build_tags(tags: list[str]) -> dict[str, semver.Version]:
    """Build the tags dictionary.

    Args:
        tags: The tags of the repository.

    Returns:
        The tags dictionary, where the key is the tag name and the value is the
            parsed tag (semantic) version, sorted by semantic version.
    """

    def tag_not_none(
        tag: tuple[str, semver.Version | None]
    ) -> TypeGuard[tuple[str, semver.Version]]:
        return tag[1] is not None

    return dict(
        sorted(
            filter(
                tag_not_none,
                ((tag, to_semver(tag)) for tag in (tags or [])),
            ),
            key=lambda tag: tag[1],
        )
    )


def _build_branches(branches: list[str]) -> dict[str, BranchVersion]:
    """Build the branches dictionary.

    Args:
        branches: The branches of the repository.

    Returns:
        The branches dictionary, where the key is the branch name and the value is
            the parsed branch name, sorted by branch version.
    """

    def branch_not_none(
        branch: tuple[str, BranchVersion | None]
    ) -> TypeGuard[tuple[str, BranchVersion]]:
        return branch[1] is not None

    return dict(
        sorted(
            filter(
                branch_not_none,
                ((branch, BranchVersion.parse(branch)) for branch in (branches or [])),
            ),
            key=lambda branch: branch[1],
        )
    )


@dataclasses.dataclass(frozen=True, kw_only=True)
class BranchVersion:
    """A branch version.

    Versions can be compared with each other. If `minor` is `None`, it is considered to
    be greater than any other `minor`.
    """

    major: int
    """The major version number."""

    minor: int | None = None
    """The minor version number."""

    name: str
    """The branch name."""

    def __post_init__(self) -> None:
        """Post init."""
        if self.major < 0:
            raise ValueError("major cannot be negative")
        if self.minor is not None and self.minor < 0:
            raise ValueError("minor cannot be negative")

    @classmethod
    def parse(cls, branch: str) -> Self | None:
        """Parse a branch name.

        See the [module documentation][frequenz.repo.config.version] for the expected
        format of the branch name.

        Args:
            branch: The branch name.

        Returns:
            The parsed branch version information or None if the branch name is not
                a valid branch name.
        """
        if match := _major_branch_re.match(branch):
            return cls(major=int(match.group(1)), name=branch)
        if match := _minor_branch_re.match(branch):
            return cls(
                major=int(match.group(1)), minor=int(match.group(2)), name=branch
            )
        _logger.debug("Invalid branch name: %s", branch)
        return None

    def __str__(self) -> str:
        """Return the branch name."""
        return self.name

    def __lt__(self, other: BranchVersion) -> bool:
        """Compare two branch version information.

        If `minor` is `None`, it is considered to be greater than any other `minor`.

        Args:
            other: The other branch version information.

        Returns:
            Whether the current branch version information is less than the other.
        """
        if not isinstance(other, BranchVersion):
            return NotImplemented
        if self.minor is None and other.minor is None:
            self_minor = 0
            other_minor = 0
        elif self.minor is None and other.minor is not None:
            self_minor = other.minor + 1
            other_minor = other.minor
        elif other.minor is None and self.minor is not None:
            self_minor = self.minor
            other_minor = self.minor + 1
        elif self.minor is not None and other.minor is not None:
            self_minor = self.minor
            other_minor = other.minor
        else:
            # We need this because mypy is not smart enough to figure it out
            assert False, "unreachable"
        return (self.major, self_minor, self.name) < (
            other.major,
            other_minor,
            other.name,
        )


class RepoVersionInfo:  # pylint: disable=too-many-instance-attributes
    """The information about a repository version.

    The information includes if it is a tag, a branch, if tags and branches are well
    formed, the next minor version, etc.

    This assumes tags follow semantic versioning and branches are in the form
    "v<major>.x.x" or "v<major>.<minor>.x", where <major> and <minor> are integers and
    represent the major and minor for the version being developed in that branch.

    New minor releases are branched from a major branch, also creating a minor branch
    for patch releases for that minor. For example, if the current major version is 1,
    the current major branch is "v1.x.x" and the current minor branch is "v1.0.x". If
    the next minor version is 1.1, the new minor branch will be "v1.1.x".
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        sha: str,
        ref: str,
        tags: list[str] | None = None,
        branches: list[str] | None = None,
    ) -> None:
        """Initialize the environment variables.

        Args:
            sha: The current commit hash.
            ref: The current reference full path (e.g. `refs/tags/v1.0.0`).
            tags: The tags of the repository.
            branches: The branches of the repository.
        """
        self._sha: str = sha
        _logger.debug("sha: %s", self._sha)

        self._ref: str = ref
        _logger.debug("ref: %s", self._ref)

        self._ref_name: str = pathlib.Path(ref).name
        _logger.debug("ref_name: %s", self._ref_name)

        self._current_branch: BranchVersion | None = BranchVersion.parse(self._ref_name)
        _logger.debug("current_branch: %s", self._current_branch)

        self._current_tag: semver.Version | None = to_semver(self._ref_name)
        _logger.debug("current_tag: %s", self._current_tag)

        self._tags: dict[str, semver.Version] = _build_tags(tags or [])
        _logger.debug("tags: %s", self._tags)

        self._branches: dict[str, BranchVersion] = _build_branches(branches or [])
        _logger.debug("branches: %s", self._branches)

    @property
    def sha(self) -> str:
        """The current commit hash."""
        return self._sha

    @property
    def ref(self) -> str:
        """The current reference full path (e.g. `refs/tags/v1.0.0`)."""
        return self._ref

    @property
    def ref_name(self) -> str:
        """The current reference name (e.g. `v1.0.0`)."""
        return self._ref_name

    @property
    def current_branch(self) -> BranchVersion | None:
        """The branch pointing to the current commit.

        `None` if there is no current branch or the name is invalid.
        """
        return self._current_branch

    @property
    def current_tag(self) -> semver.Version | None:
        """The tag pointing to the current commit.

        `None` if there is no tag or the name is invalid.
        """
        return self._current_tag

    @property
    def tags(self) -> dict[str, semver.Version]:
        """The tags of the repository.

        The key is the tag name and the value is the parsed tag (semantic) version.
        """
        return self._tags

    @property
    def branches(self) -> dict[str, BranchVersion]:
        """The branches of the repository.

        The key is the branch name and the value is the parsed branch name.
        """
        return self._branches

    def find_last_tag(self) -> semver.Version | None:
        """Find the last tag.

        Returns:
            If we are at a tag, return the [current
                tag][frequenz.repo.config.version.RepoVersionInfo.current_tag]. If we
                are at a branch, return the last tag matching the branch major and minor
                (if any). If there are no matching tags, return `None`.
        """
        if self._current_tag:
            return self._current_tag
        branch = self.current_branch
        if branch is None:
            return None
        tags = [t for t in self._tags.values() if t.major == branch.major]
        if branch.minor is not None:
            tags = [t for t in tags if t.minor == branch.minor]
        if not tags:
            return None
        return max(tags)

    def find_next_breaking_branch(self) -> BranchVersion | None:
        """Find the next branch potentially introducing breaking changes.

        Returns:
            If there is a last tag, use that as a base, otherwise use the current branch
                as a base. If none is available, return `None`. If there is a base, the
                major is incremented by one and the minor is set to `None`, unless the
                major is 0 (an "initial development version" for semver), in which case
                the major is set to 0 and the minor is incremented by one, as the next
                minor could be a breaking change. Technically semver allows breaking
                changes in patches for major version 0, but we assume patches maintain
                backwards compatibility.
        """
        v_prefix = "v" if self._ref_name.startswith("v") else ""
        last_tag = self.find_last_tag()
        if last_tag is None:
            branch = self.current_branch
            if branch is None:
                _logger.warning(
                    "Trying to get the next breaking branch but there is no (valid) "
                    "last tag nor current branch for %r",
                    self._ref,
                )
                return None
            major = branch.major + 1
            minor = branch.minor or 1
        else:
            major = last_tag.major + 1
            minor = last_tag.minor + 1

        # If the next major is 1, then the current is 0, so the next minor could be
        # a breaking change.
        if major == 1:
            return BranchVersion(major=0, minor=minor, name=f"{v_prefix}0.{minor}.x")

        return BranchVersion(major=major, name=f"{v_prefix}{major}.x.x")

    def find_next_minor_for_major_branch(self) -> int | None:
        """Find the next minor version for the current major branch.

        Returns:
            The next minor version or `None` if there is no current branch, the
                current branch is invalid or is not a major branch.
        """
        branch = self.current_branch
        if branch is None:
            _logger.warning(
                "Trying to get the next minor for the current branch but "
                "the current branch is invalid: %r",
                branch,
            )
            return None

        if branch.minor is not None:
            _logger.warning(
                "Trying to get the next minor for the current branch but "
                "the current branch is not a major branch: %r",
                branch,
            )
            return None

        minor_tags = [
            tag
            for tag in self._tags.values()
            if tag.major == branch.major and tag.prerelease is None
        ]
        if not minor_tags:
            return 0

        bigger_tag = max(minor_tags, key=lambda tag: tag.minor)
        return bigger_tag.minor + 1

    def is_tag(self) -> bool:
        """Tell whether we are at a stable release."""
        return self._ref.startswith("refs/tags/")

    def is_branch(self) -> bool:
        """Tell whether we are at a major branch."""
        return self._ref.startswith("refs/heads/")

    def is_tag_last_minor_for_major(self) -> bool:
        """Tell whether the current tag is the last minor version for the major version.

        If we are not at a tag or there are not tags in the repo, return `False`.

        If the `current_tag` is a pre-release, only pre-release tags are considered and
        if is not a pre-release, only stable tags are considered.
        """
        tag = self.current_tag
        if tag is None:
            _logger.warning(
                "Can't determine if %r is the last minor tag for the major version "
                "because we are not at a tag or the tag is invalid",
                self._ref_name,
            )
            return False
        minor_tags = [tag_ for tag_ in self._tags.values() if tag_.major == tag.major]
        if tag.prerelease is None:
            minor_tags = [tag_ for tag_ in minor_tags if tag_.prerelease is None]
        else:
            minor_tags = [tag_ for tag_ in minor_tags if tag_.prerelease is not None]
        if not minor_tags:
            _logger.debug(
                "No minor tags found for major version %s, considering the "
                "tag %r the last minor for this major",
                tag.major,
                tag,
            )
            return True
        return tag >= max(minor_tags, key=lambda tag_: tag_.minor)

    def is_tag_latest(self) -> bool:
        """Tell whether the current tag is the latest tag.

        The latest tag is the tag with the biggest major, minor and patch
        version. If the current tag is a prerelease, then only prereleases are used to
        determine the biggest.

        Sorting is always according to semver.
        """
        tag = self.current_tag
        if tag is None:
            _logger.warning(
                "Can't determine if %r is the latest tag because we are not at a tag "
                "or the tag is invalid",
                self._ref_name,
            )
            return False
        if tag.prerelease is None:
            tags_it = (tag_ for tag_ in self._tags.values() if tag_.prerelease is None)
        else:
            tags_it = (
                tag_ for tag_ in self._tags.values() if tag_.prerelease is not None
            )
        latest = sorted(tags_it, reverse=True)
        if not latest:
            _logger.warning(
                "No other tags found, at least the current tag %r should be in the list "
                "of tags (%r)",
                tag,
                self._tags,
            )
            return False
        return tag == latest[0]

    def is_branch_latest(self) -> bool:
        """Tell whether the current branch is the latest.

        The latest branch is the branch with the biggest major and minor, but if minor
        is `None`, then it is considered the biggest minor.
        """
        branch = self.current_branch
        if branch is None:
            _logger.warning(
                "Can't determine if %r is the latest branch because we are not at "
                "a branch or the branch is invalid",
                self._ref_name,
            )
            return False
        latest = sorted(self._branches.values(), reverse=True)
        if not latest:
            _logger.warning(
                "No other branches found, at least the current branch %r should be in the list "
                "of branches (%r)",
                branch,
                self._branches,
            )
            return False
        return branch == latest[0]



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Tools to work with mike.

This module provides these tools:

* Building the mike version information from the repository information
  ([`build_mike_version()`][frequenz.repo.config.mkdocs.mike.build_mike_version]).
* Sorting the mike version information `version.json` file
  ([`sort_mike_versions()`][frequenz.repo.config.mkdocs.mike.sort_mike_versions]).
* Comparing mike versions
  ([`compare_mike_version()`][frequenz.repo.config.mkdocs.mike.compare_mike_version]).

Mike versions have the format `vX.Y(-pre|-dev)?`, where `X` is the major version, `Y` is
the minor version, and the optional suffix is either `-pre` for pre-release versions or
`-dev` for development versions.

Stable (tagged) versions have the format `vX.Y`, unless they are pre-release versions,
then they have the format `vX.Y-pre`. Develoment branches have the format `vX.Y-dev`.

* Tag `v1.0.0` -> `v1.0`
* Tag `v2.1.0-alpha.1` -> `v2.1-pre`
* Branch `v1.x.x` (with no releases) -> `v1.0-dev`
* Branch `v1.x.x (with releases an existing release, for example `v1.0.0`) -> `v1.1-dev`
* Branch `v1.1.x` -> `v1.1-dev`

Aliases have the format `vX(-pre|-dev)?`, where `X` is the major version and the
optional suffix is either `-pre` for pre-release versions or `-dev` for development
versions. An alias is also provided to point to the latest version, which is `latest`
for stable versions, `latest-pre` for pre-release versions, and `latest-dev` for
development versions.
"""

import dataclasses
import functools
import re

import semver

from ..version import RepoVersionInfo


@dataclasses.dataclass(frozen=True, kw_only=True)
class MikeVersionInfo:
    """The information needed to publish a mike version.

    This is what mike needs when publishing documentation for a particular version.
    """

    version: str
    """The version identifier of this version."""

    title: str = ""
    """The title of this version."""

    aliases: list[str] = dataclasses.field(default_factory=list)
    """The aliases of this version."""


def build_mike_version(repo_info: RepoVersionInfo) -> MikeVersionInfo:
    """Build the mike version information from the given repository information.

    The version is build based on if a tag or a branch is checked out.

    For tags, the title is the tag name, the version is "vX.Y", where X is the major
    version and Y is the minor version of the tag. If the tag is the last minor version
    for the major version, the alias "vX" is added. If the tag is the latest tag, the
    alias "latest" is added.

    For pre-release tags it's the same but the "-pre" suffix is added to the version and
    the aliases.

    For branches, the title is "vX.Y-dev (SHA)", where X is the major version and Y is
    the minor version of the branch. The version is "vX.Y-dev". If the branch is the
    latest branch, the alias "latest-dev" is added.

    Args:
        repo_info: The repository information.

    Returns:
        The mike version.

    Raises:
        ValueError: If the given repository information is invalid or versions can't be
            determined for some other reason.
    """
    title: str = ""
    version: str = ""
    aliases: list[str] = []

    if repo_info.is_tag():
        current_tag = repo_info.current_tag
        if current_tag is None:
            raise ValueError(
                f"The tag {repo_info.ref_name!r} is not a valid semver version",
            )
        suffix = ""
        if current_tag.prerelease is not None:
            suffix = "-pre"
        title = repo_info.ref_name  # vX.Y.Z
        version = f"v{current_tag.major}.{current_tag.minor}{suffix}"  # vX.Y(-pre)?
        if repo_info.is_tag_last_minor_for_major():
            aliases.append(f"v{current_tag.major}{suffix}")  # vX(-pre)?
        if repo_info.is_tag_latest():
            aliases.append(f"latest{suffix}")  # latest(-pre)?
        return MikeVersionInfo(title=title, version=version, aliases=aliases)

    if repo_info.is_branch():
        current_branch = repo_info.current_branch
        if current_branch is None:
            raise ValueError(
                f"The branch {repo_info.ref_name!r} is not a valid branch name",
            )
        minor: int
        if current_branch.minor is None:
            next_minor = repo_info.find_next_minor_for_major_branch()
            if next_minor is None:
                raise ValueError(
                    f"Could not determine the next minor version for {current_branch.name!r}",
                )
            minor = next_minor
            aliases.append(f"v{current_branch.major}-dev")  # vX-dev
        else:
            minor = current_branch.minor

        major = current_branch.major
        title = f"v{major}.{minor}-dev ({repo_info.sha[:7]})"
        version = f"v{major}.{minor}-dev"  # vX.Y-dev
        if repo_info.is_branch_latest():
            aliases.append("latest-dev")
        return MikeVersionInfo(title=title, version=version, aliases=aliases)

    raise ValueError(
        f"Don't know how to handle '{repo_info.ref}' to make 'mike' version",
    )


_is_version_re = re.compile(r"^v(\d+).(\d+)(-dev|-pre)?$")
_stable_to_semver_re = re.compile(r"^v(\d+).(\d+)$")
_pre_to_semver_re = re.compile(r"^v(\d+).(\d+)-pre$")
_dev_to_semver_re = re.compile(r"^v(\d+).(\d+)-dev$")


def _to_fake_sortable_semver(version: str) -> str:
    """Convert a branch version string to a semver string.

    The following transformations are applied:

    - `vX.Y-pre` -> `X.Y.0-pre`
    - `vX.Y`     -> `X.Y.0`
    - `vX.Y-dev` -> `X.Y.999999`

    The idea is to convert the version string to a semver string that can be sorted
    together with proper semver tags using the semver built-in sorting.

    Args:
        version: The version string to convert.

    Returns:
        The converted version string.
    """
    version = _stable_to_semver_re.sub(r"\1.\2.0", version)
    version = _pre_to_semver_re.sub(r"\1.\2.0-pre", version)
    version = _dev_to_semver_re.sub(r"\1.\2.999999", version)
    if version.startswith("v"):
        version = version[1:]
    return version


def compare_mike_version(version1: str, version2: str) -> int:
    """Compare two versions.

    The versions are compared as follows:

    - Versions are first compared by major version (`X`).
    - If they have the same major, then they are compared by minor version (`Y`).
    - If they have the same major and minor, then stable versions (`vX.Y`) are
      considered bigger than pre-releases (`vX.Y-pre`) and development versions
      (`vX.Y-dev`) are considered bigger than pre-releases.
    - Any other version not matching `vX.Y(-pre|-dev)?` is considered to be bigger than
      the matching versions.
    - Not matching versions are compared alphabetically.

    Example:

        `v1.0-pre` < `v1.0` < `v1.0-dev` < `v1.1` < `v2.0-pre` < `v2.0` < `v2.0-dev`
        < `whatever` < `x`.

    Args:
        version1: The first version to compare.
        version2: The second version to compare.

    Returns:
        A negative number if `version1` is older than `version2`, a positive number if
            `version1` is newer than `version2`, or zero if they are equal.
    """
    is_version_v1 = _is_version_re.match(version1)
    is_version_v2 = _is_version_re.match(version2)
    if is_version_v1 and is_version_v2:
        return semver.Version.parse(_to_fake_sortable_semver(version1)).compare(
            _to_fake_sortable_semver(version2)
        )

    if is_version_v1:  # version2 is not a version
        return -1
    if is_version_v2:  # version1 is not a version
        return 1

    return -1 if version1 < version2 else 1


def sort_mike_versions(versions: list[str], *, reverse: bool = True) -> list[str]:
    """Sort `mike`'s `version.json` file with a custom order.

    The `version` keys are expected as follows:

    - `vX.Y` for stable release versions
    - `vX.Y-pre` for pre-release versions
    - `vX.Y-dev` for development versions
    - Any other arbitrary string for other versions

    The sorting order is as follows:

    - Versions are first sorted by major version (`X`).
    - Inside a major version group, versions are sorted by minor version (`Y`).
    - For the same major and minor version, development versions (`-dev`) considered
      the latest for that major version group, then stable versions, and finally
      pre-release versions (`-pre`).
    - Other versions appear first and are sorted alphabetically.

    The versions are sorted in-place using
    [`compare_mike_version()`][frequenz.repo.config.mkdocs.mike.compare_mike_version].

    Example:

        `z`, `whatever`, `v2.1-dev`, `v2.1`, `v2.1-pre`, `v2.0`, `v1.1-dev`, `v1.0-dev`,
        `v1.0`

    Args:
        versions: The list of versions to sort.
        reverse: Whether to sort in reverse order.

    Returns:
        The sorted list of versions.
    """
    versions.sort(key=functools.cmp_to_key(compare_mike_version), reverse=reverse)
    return versions



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Sort `mike`'s `version.json` file with a custom order."""


import json
import sys
from typing import Any, TextIO

from .... import github
from ....mkdocs.mike import sort_mike_versions


def _load_and_sort_versions_from(stream: TextIO) -> dict[str, dict[str, Any]]:
    """Load the versions from the given stream and sort them.

    Args:
        stream: The stream to read the versions from.

    Returns:
        The sorted loaded versions.
    """
    versions = {v["version"]: v for v in json.load(stream)}
    return {v: versions[v] for v in sort_mike_versions(list(versions.keys()))}


def _dump_versions_to(versions: dict[str, dict[str, Any]], stream: TextIO) -> None:
    """Dump the versions to the given stream.

    Args:
        versions: The versions to dump.
        stream: The stream to write the versions to.
    """
    json.dump(list(versions.values()), stream, separators=(",", ":"))


def main() -> None:
    """Sort `mike`'s `version.json` file with a custom order.

    The versions are sorted using `sort_versions()`.

    If no arguments are given, then the contents are read from stdin and the sorted
    versions are printed to stdout.

    If one argument is given, then the contents of the file are replaced with the sorted
    versions.

    If more than one argument is given, then an error is printed to stderr and the
    program exits with a non-zero exit code.
    """
    github.configure_logging()

    match len(sys.argv):
        case 1:
            _dump_versions_to(_load_and_sort_versions_from(sys.stdin), sys.stdout)
        case 2:
            with open(sys.argv[1], "r", encoding="utf8") as stream_in:
                versions = _load_and_sort_versions_from(stream_in)
            with open(sys.argv[1], "w", encoding="utf8") as stream_out:
                _dump_versions_to(versions, stream_out)

        case _:
            print(
                f"""\
Usage: {sys.argv[0]} [<versions.json>]

If <versions.json> is given, the contents will be replaced with the sorted versions.
Otherwise, the contents are read from stdin and the sorted versions are printed to stdout.
""",
                file=sys.stderr,
            )
            sys.exit(2)


if __name__ == "__main__":
    main()



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Utilities to work with GitHub.

- [`abort()`][frequenz.repo.config.github.abort] to print an error message using GitHub
    Actions commands and exit.
- [`require_env()`][frequenz.repo.config.github.require_env] to get an environment
    variable or exit with an error if it is not defined.
- [`get_tags()`][frequenz.repo.config.github.get_tags] to get the tags of a repository.
- [`get_branches()`][frequenz.repo.config.github.get_branches] to get the branches of a
    repository.
- [`configure_logging()`][frequenz.repo.config.github.configure_logging] to configure
    logging for GitHub Actions.
"""


import logging
import os
import subprocess
import sys
from typing import NoReturn

import github_action_utils as gha

from . import version

_logger = logging.getLogger(__name__)


class GitHubActionsFormatter(logging.Formatter):
    """A formatter for GitHub Actions."""

    level_mapping: dict[int, str] = {
        logging.CRITICAL: "error",
        logging.ERROR: "error",
        logging.WARNING: "warning",
        logging.INFO: "notice",
        logging.DEBUG: "debug",
    }
    """The mapping from logging levels to GitHub Actions levels."""

    def format(self, record: logging.LogRecord) -> str:
        """Format the record using GitHub Actions commands.

        Args:
            record: The record to format.

        Returns:
            The formatted record.
        """
        github_level = self.level_mapping.get(record.levelno, "notice")
        github_command = f"::{github_level}::{record.getMessage()}"
        return github_command


def abort(  # pylint: disable=too-many-arguments
    message: str,
    title: str | None = None,
    file: str | None = None,
    col: int | None = None,
    end_column: int | None = None,
    line: int | None = None,
    end_line: int | None = None,
    use_subprocess: bool = False,
    error_code: int = 1,
) -> NoReturn:
    """Print an error message using GitHub Actions commands and exit.

    Args:
        message: The message of the error.
        title: The title of the error.
        file: The file where the error occurred.
        col: The column where the error occurred.
        end_column: The end column where the error occurred.
        line: The line where the error occurred.
        end_line: The end line where the error occurred.
        use_subprocess: Whether to use subprocess to print the error.
        error_code: The error code to exit with.
    """
    gha.error(
        message,
        title=title,
        file=file,
        col=col,
        end_column=end_column,
        line=line,
        end_line=end_line,
        use_subprocess=use_subprocess,
    )
    sys.exit(error_code)


def require_env(name: str) -> str:
    """Get the environment variable.

    Args:
        name: The name of the environment variable.

    Returns:
        The environment variable.

    Raises:
        ValueError: If the environment variable is not defined.
    """
    value = os.environ.get(name)
    if value is None:
        raise ValueError(f"Environment variable {name!r} is not defined")
    return value


def get_tags(repository: str) -> list[str]:
    """Get the tags of the repository.

    This function uses the `TAGS` environment variable if it is defined. If it is
    not defined, it uses the GitHub `gh` CLI tool to get them. This means the tool
    needs to be properly configured and have at least read access over
    `repository`.

    Args:
        repository: The repository to get the tags of.

    Returns:
        The tags of the repository.
    """
    tags_str: list[str]
    if env_tags := os.environ.get("TAGS", None):
        tags_str = env_tags.split()
    else:
        tags_str = (
            subprocess.check_output(
                [
                    "gh",
                    "api",
                    "-q",
                    ".[].name",
                    "-H",
                    "Accept: application/vnd.github+json",
                    "-H",
                    "X-GitHub-Api-Version: 2022-11-28",
                    f"/repos/{repository}/tags",
                ]
            )
            .decode("utf-8")
            .splitlines()
        )

    _logger.debug("Got tags: %r", tags_str)
    return tags_str


def get_branches(repository: str) -> list[str]:
    """Get the branches of the repository.

    This function uses the `BRANCHES` environment variable if it is defined. If it
    is not defined, it uses the GitHub `gh` CLI tool to get them. This means the
    tool needs to be properly configured and have at least read access over
    `repository`.

    Args:
        repository: The repository to get the branches of.

    Returns:
        The branches of the repository.
    """
    branches_str: list[str]
    if env_branches := os.environ.get("BRANCHES", None):
        branches_str = env_branches.split()
    else:
        branches_str = (
            subprocess.check_output(
                [
                    "gh",
                    "api",
                    "-q",
                    ".[].name",
                    "-H",
                    "Accept: application/vnd.github+json",
                    "-H",
                    "X-GitHub-Api-Version: 2022-11-28",
                    f"/repos/{repository}/branches",
                ]
            )
            .decode("utf-8")
            .splitlines()
        )
    _logger.debug("Got branches: %r", branches_str)
    return branches_str


def configure_logging(level: int | None = None) -> None:
    """Configure logging for GitHub Actions.

    The
    [`GitHubActionsFormatter`][frequenz.repo.config.github.GitHubActionsFormatter]
    is used to format the logs, using [GitHub Action
    commands](https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions).

    Args:
        level: The logging level to use. If `None`, the level is set to
            `logging.INFO` unless the `RUNNER_DEBUG` environment variable
            is set to `1`, in which case the level is set to
            `logging.DEBUG`.
    """
    # Just for checking it is defined, github_action_utils deals with it automatically.
    require_env("GITHUB_OUTPUT")
    is_debug_run = os.environ.get("RUNNER_DEBUG", None) == "1"
    handler = logging.StreamHandler()
    handler.setFormatter(GitHubActionsFormatter())
    if level is None:
        level = logging.DEBUG if is_debug_run else logging.INFO
    logging.basicConfig(level=level, handlers=[handler])


def get_repo_version_info() -> version.RepoVersionInfo:
    """Get the repository version information.

    This function uses the `GITHUB_REPO`, `GIT_REF`, and `GIT_SHA` environment
    variables to get the repository information. If these variables are not
    defined, it raises a `ValueError`.

    It also uses the `BRANCHES` and `TAGS` environment variables to get the
    branches and tags of the repository. If they are not defined, the GitHub `gh`
    CLI tool is used to get them. This means the tool needs to be properly
    configured and have at least read access over `repository`.

    Returns:
        The repository version information.
    """
    repository = require_env("GITHUB_REPO")
    repo_info = version.RepoVersionInfo(
        ref=require_env("GIT_REF"),
        sha=require_env("GIT_SHA"),
        tags=get_tags(repository),
        branches=get_branches(repository),
    )
    return repo_info



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Pytest utilities.

This package contains utilities for testing with [`pytest`](https://pypi.org/project/pytest/).

The following modules are available:

- [`examples`][frequenz.repo.config.pytest.examples]: Utilities to enable linting of
  code examples in docstrings.
"""



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Tools to manage and generate documentation using `mkdocs`."""

from .api_pages import generate_protobuf_api_pages, generate_python_api_pages

__all__ = [
    "generate_protobuf_api_pages",
    "generate_python_api_pages",
]



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Utility to enable linting of code examples in docstrings.

Code examples are often wrapped in triple backticks (````python`) within our docstrings.
This plugin extracts these code examples and validates them using pylint.

The main utility function is
[`get_sybil_arguments()`][frequenz.repo.config.pytest.examples.get_sybil_arguments],
which returns a dictionary that can be used to pass to the [`Sybil()`][sybil.Sybil]
constructor.

You still need to create a `conftest.py` file in the root of your project's sources,
typically `src/conftest.py`, with the following contents:

```python
from frequenz.repo.config.pytest import examples
from sybil import Sybil

pytest_collect_file = Sybil(**examples.get_sybil_arguments()).pytest()
```
"""

import ast
import os
import subprocess
from pathlib import Path
from typing import Any

from sybil import Example
from sybil.evaluators.python import pad
from sybil.parsers.abstract.lexers import textwrap
from sybil.parsers.myst import CodeBlockParser

_PYLINT_DISABLE_COMMENT = (
    "# pylint: {}=unused-import,wildcard-import,unused-wildcard-import"
)

_FORMAT_STRING = """
# Generated auto-imports for code example
{disable_pylint}
{imports}
{enable_pylint}

{code}"""


def get_sybil_arguments() -> dict[str, Any]:
    """Get the arguments to pass when instantiating the Sybil object to lint docs examples.

    Returns:
        The arguments to pass when instantiating the Sybil object.
    """
    return {
        "parsers": [_CustomPythonCodeBlockParser()],
        "patterns": ["*.py"],
        # This is a hack because Sybil seems to have issues with `__init__.py` files.
        # See https://github.com/frequenz-floss/frequenz-repo-config-python/issues/113
        # for details
        "excludes": ["__init__.py"],
    }


def _get_import_statements(code: str) -> list[str]:
    """Get all import statements from a given code string.

    Args:
        code: The code to extract import statements from.

    Returns:
        A list of import statements.
    """
    tree = ast.parse(code)
    import_statements: list[str] = []

    for node in ast.walk(tree):
        if isinstance(node, (ast.Import, ast.ImportFrom)):
            import_statement = ast.get_source_segment(code, node)
            assert import_statement is not None
            import_statements.append(import_statement)

    return import_statements


def _path_to_import_statement(path: Path) -> str:
    """Convert a path to a Python file to an import statement.

    Args:
        path: The path to convert.

    Returns:
        The import statement.

    Raises:
        ValueError: If the path does not point to a Python file.
    """
    # Make the path relative to the present working directory
    if path.is_absolute():
        path = path.relative_to(Path.cwd())

    # Check if the path is a Python file
    if path.suffix != ".py":
        raise ValueError("Path must point to a Python file (.py)")

    # Remove 'src' prefix if present
    parts = path.parts
    if parts[0] == "src":
        parts = parts[1:]

    # Remove the '.py' extension and join parts with '.'
    module_path = ".".join(parts)[:-3]

    # Create the import statement
    import_statement = f"from {module_path} import *"
    return import_statement


# We need to add the type ignore comment here because the Sybil library does not
# have type annotations.
class _CustomPythonCodeBlockParser(CodeBlockParser):  # type: ignore[misc]
    """Code block parser that validates extracted code examples using pylint.

    This parser is a modified version of the default Python code block parser
    from the Sybil library.
    It uses pylint to validate the extracted code examples.

    All code examples are preceded by the original file's import statements as
    well as an wildcard import of the file itself.
    This allows us to use the code examples as if they were part of the original
    file.

    Additionally, the code example is padded with empty lines to make sure the
    line numbers are correct.

    Pylint warnings which are unimportant for code examples are disabled.
    """

    def __init__(self) -> None:
        """Initialize the parser."""
        super().__init__("python")

    def evaluate(self, example: Example) -> None | str:
        """Validate the extracted code example using pylint.

        Args:
            example: The extracted code example.

        Returns:
            None if the code example is valid, otherwise the pylint output.
        """
        # Get the import statements for the original file
        import_header = _get_import_statements(example.document.text)
        # Add a wildcard import of the original file
        import_header.append(
            _path_to_import_statement(Path(os.path.relpath(example.path)))
        )
        imports_code = "\n".join(import_header)

        # Dedent the code example
        # There is also example.parsed that is already prepared, but it has
        # empty lines stripped and thus fucks up the line numbers.
        example_code = textwrap.dedent(
            example.document.text[example.start : example.end]
        )
        # Remove first line (the line with the triple backticks)
        example_code = example_code[example_code.find("\n") + 1 :]

        example_with_imports = _FORMAT_STRING.format(
            disable_pylint=_PYLINT_DISABLE_COMMENT.format("disable"),
            imports=imports_code,
            enable_pylint=_PYLINT_DISABLE_COMMENT.format("enable"),
            code=example_code,
        )

        # Make sure the line numbers are correct
        source = pad(
            example_with_imports,
            example.line - imports_code.count("\n") - _FORMAT_STRING.count("\n"),
        )

        # pylint disable parameters
        pylint_disable_params = [
            "missing-module-docstring",
            "missing-class-docstring",
            "missing-function-docstring",
            "reimported",
            "unused-variable",
            "no-name-in-module",
            "await-outside-async",
        ]

        response = _validate_with_pylint(source, example.path, pylint_disable_params)

        if len(response) > 0:
            return (
                f"Pylint validation failed for code example:\n"
                f"{example_with_imports}\nOutput: " + "\n".join(response)
            )

        return None


def _validate_with_pylint(
    code_example: str, path: str, disable_params: list[str]
) -> list[str]:
    """Validate a code example using pylint.

    Args:
        code_example: The code example to validate.
        path: The path to the original file.
        disable_params: The pylint disable parameters.

    Returns:
        A list of pylint messages.
    """
    try:
        pylint_command = [
            "pylint",
            "--disable",
            ",".join(disable_params),
            "--from-stdin",
            path,
        ]

        subprocess.run(
            pylint_command,
            input=code_example,
            text=True,
            capture_output=True,
            check=True,
        )
    except subprocess.CalledProcessError as exception:
        output = exception.output
        assert isinstance(output, str)
        return output.splitlines()

    return []



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Command-line tool to get the current `mike` version information of a repository.

For now this tool is designed to be used in GitHub Actions workflows, but is is possible
to use it in other environments as well. To do so the `gh` tool should be properly
configured and have at least read-access to the repository, and the following environment
variables should be set:

- `GITHUB_REPO`: The repository to get the version information of (e.g.
  `frequenz-floss/frequenz-sdk-python`).
- `GIT_REF`: The git reference to get the version information of (e.g.
  `refs/heads/v1.x.x`).
- `GIT_SHA`: The git sha to get the version information of (e.g.
  `84df6ad1d9990d7afd47a9f8e8a386702b09eba0`).
- `TAGS`: The tags of the repository (e.g. `v1.0.0 v1.0.1`).
- `BRANCHES`: The branches of the repository (e.g. `v1.x.x v2.x.x`).
- `GITHUB_OUTPUT`: The output variable to set the version information to (e.g.
  `/dev/stdout`).
"""

import logging

import github_action_utils as gha

from .... import github
from ....mkdocs import mike

_logger = logging.getLogger(__name__)


def _output_gha_vars(mike_version: mike.MikeVersionInfo) -> None:
    """Output mike version information as GitHub Action variables."""
    _logger.debug("Outputting mike version information as GitHub Action variables:")
    _logger.debug("  title: %r", mike_version.title)
    _logger.debug("  version: %r", mike_version.version)
    _logger.debug("  aliases: %r", " ".join(mike_version.aliases))

    gha.set_output("title", mike_version.title)
    gha.set_output("version", mike_version.version)
    gha.set_output("aliases", " ".join(mike_version.aliases))


def main() -> None:
    """Output mike version variables for GitHub Actions."""
    github.configure_logging()

    try:
        mike_version = mike.build_mike_version(github.get_repo_version_info())
    except ValueError as error:
        github.abort(f"{error}.", title="Documentation was not published")

    _output_gha_vars(mike_version)


if __name__ == "__main__":
    main()
